@@ -1,6 +1,10 @@
+"""
+Module containing the CNLP transformer model.
+"""
 # from transformers.models.auto import  AutoModel, AutoConfig
 import copy
-from typing import Optional, List
+import inspect
+from typing import Optional, List, Any, Dict
 
 from transformers import AutoModel, AutoConfig
 from transformers.modeling_utils import PreTrainedModel
@@ -97,10 +101,37 @@ class RepresentationProjectionLayer(nn.Module):
 
 
 class CnlpConfig(PretrainedConfig):
+    """
+    The config class for :class:`CnlpModelForClassification`.
+
+    :param encoder_name: the encoder name to use with :meth:`transformers.AutoConfig.from_pretrained`
+    :param typing.Optional[str] finetuning_task: the tasks for which this model is fine-tuned
+    :param typing.List[int] num_labels_list: the number of labels for each task
+    :param int layer: the index of the encoder layer to extract features from
+    :param bool tokens: if true, sentence-level classification is done based on averaged token embeddings for token(s) surrounded by <e> </e> special tokens
+    :param int num_rel_attention_heads: the number of features/attention heads to use in the NxN relation classifier
+    :param int rel_attention_head_dims: the number of parameters in each attention head in the NxN relation classifier
+    :param typing.List[bool] tagger: for each task, whether the task is a sequence tagging task
+    :param typing.List[bool] relations: for each task, whether the task is a relation extraction task
+    :param bool use_prior_tasks: whether to use the outputs from the previous tasks as additional inputs for subsequent tasks
+    :param \**kwargs: arguments for :class:`transformers.PretrainedConfig`
+    """
     model_type='cnlpt'
 
-    def __init__(self, encoder_name='roberta-base', finetuning_task=None, num_labels_list=[], layer=-1, tokens=False, num_rel_attention_heads=12, rel_attention_head_dims=64, tagger = [False], relations = [False], use_prior_tasks=False, num_tokens=-1,
-    **kwargs):
+    def __init__(
+        self,
+        encoder_name='roberta-base',
+        finetuning_task=None,
+        num_labels_list=[],
+        layer=-1,
+        tokens=False,
+        num_rel_attention_heads=12,
+        rel_attention_head_dims=64,
+        tagger = [False],
+        relations = [False],
+        use_prior_tasks=False,
+        **kwargs
+     ):
         super().__init__(**kwargs)
         # self.name_or_path='cnlpt'
         self.finetuning_task = finetuning_task
@@ -114,18 +145,34 @@ class CnlpConfig(PretrainedConfig):
         self.use_prior_tasks = use_prior_tasks
         self.encoder_name = encoder_name
         self.encoder_config = AutoConfig.from_pretrained(encoder_name).to_dict()
-        try:
-            self.hidden_dropout_prob = self.encoder_config['hidden_dropout_prob']
-            self.hidden_size = self.encoder_config['hidden_size']
-        except KeyError as ke:
-            raise ValueError(f'Encoder config does not have an attribute "{ke.args[0]}";'
-                             f' this is likely because the API of the chosen encoder'
-                             f' differs from the BERT/RoBERTa API (e.g. with DistilBERT).'
-                             f' Encoders with different APIs are not yet supported (#35).')
-        self.num_tokens = num_tokens
+        if encoder_name.startswith('distilbert'):
+            self.hidden_dropout_prob = self.encoder_config['dropout']
+            self.hidden_size = self.encoder_config['dim']
+        else:
+            try:
+                self.hidden_dropout_prob = self.encoder_config['hidden_dropout_prob']
+                self.hidden_size = self.encoder_config['hidden_size']
+            except KeyError as ke:
+                raise ValueError(f'Encoder config does not have an attribute'
+                                 f' "{ke.args[0]}"; this is likely because the API of'
+                                 f' the chosen encoder differs from the BERT/RoBERTa'
+                                 f' API and the DistilBERT API. Encoders with different'
+                                 f' APIs are not yet supported (#35).')
 
 
 class CnlpModelForClassification(PreTrainedModel):
+    """
+    The CNLP transformer model.
+
+    :param typing.Optional[typing.List[float]] class_weights: if provided,
+        the weights to use for each task when computing the loss
+    :param float final_task_weight: the weight to use for the final task
+        when computing the loss; default 1.0.
+    :param float argument_regularization: if provided, the argument
+        regularization to use when computing the loss
+    :param bool freeze: whether to freeze the weights of the encoder
+    :param bool bias_fit: whether to fine-tune only the bias of the encoder
+    """
     base_model_prefix = 'cnlpt'
     config_class = CnlpConfig
 
@@ -148,6 +195,11 @@ class CnlpModelForClassification(PreTrainedModel):
         self.encoder = encoder_model.from_pretrained(config.encoder_name)
         self.encoder.resize_token_embeddings(encoder_config.vocab_size)
 
+        if config.layer > len(encoder_model.encoder.layer):
+            raise ValueError('The layer specified (%d) is too big for the specified encoder which has %d layers' % (
+                config.layer,
+                len(encoder_model.encoder.layer)
+            ))
         self.num_labels = config.num_labels_list
         
         if freeze:
@@ -232,6 +284,7 @@ class CnlpModelForClassification(PreTrainedModel):
             batch_size:
             seq_len:
             state:
+        :meta private:
         """
         if task_num_labels == 1:
             #  We are doing regression
@@ -288,6 +341,7 @@ class CnlpModelForClassification(PreTrainedModel):
             logits (`List[torch.FloatTensor]`): the computed logits of the batch.
             attention_mask (`torch.LongTensor`): the attention mask for the batch.
             state: the state dict containing the loss for the batch.
+        :meta private:
         """
         # standard e2e relation task -- two entity extractors and relation extractor.
         prob_no_rel = softmax(logits[2], dim=3)[:, :, :, 0]
@@ -332,6 +386,20 @@ class CnlpModelForClassification(PreTrainedModel):
 
         state["loss"] += self.argument_regularization * prob_rel_no_ent.sum()
 
+    def generalize_encoder_forward_kwargs(self, **kwargs: Any) -> Dict[str, Any]:
+        new_kwargs = dict()
+        params = inspect.signature(self.encoder.forward).parameters
+        for name, value in kwargs.items():
+            if name not in params and value is not None:
+                # Warn if a contentful parameter is not valid
+                logger.warning(f"Parameter {name} not present for encoder class {self.encoder.__class__.__name__}.")
+            elif name in params:
+                # Pass all, and only, parameters that are valid,
+                # regardless of whether they are None
+                new_kwargs[name] = value
+            # else, value is None and not in params, so we ignore it
+        return new_kwargs
+
     def forward(
         self,
         input_ids=None,
@@ -346,15 +414,34 @@ class CnlpModelForClassification(PreTrainedModel):
         event_tokens=None,
     ):
         r"""
-        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size,)`, `optional`, defaults to :obj:`None`):
-            Labels for computing the sequence classification/regression loss.
-            Indices should be in :obj:`[0, ..., config.num_labels - 1]`.
-            If :obj:`config.num_labels == 1` a regression loss is computed (Mean-Square loss),
-            If :obj:`config.num_labels > 1` a classification loss is computed (Cross-Entropy).
+        Forward method.
+
+        Args:
+            input_ids (`torch.LongTensor` of shape `(batch_size, sequence_len)`, *optional*):
+                A batch of chunked documents as tokenizer indices.
+            attention_mask (`torch.LongTensor` of shape `(batch_size, sequence_len)`, *optional*):
+                Attention masks for the batch.
+            token_type_ids (`torch.LongTensor` of shape `(batch_size, sequence_len)`, *optional*):
+                Token type IDs for the batch.
+            position_ids: (`torch.LongTensor` of shape `(batch_size, sequence_len)`, *optional*):
+                Position IDs for the batch.
+            head_mask (`torch.LongTensor` of shape `(num_heads,)` or `(num_layers, num_heads)`, *optional*):
+                Token encoder head mask.
+            inputs_embeds (`torch.FloatTensor` of shape `(batch_size, sequence_len, hidden_size)`, *optional*):
+                A batch of chunked documents as token embeddings.
+            labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
+                Labels for computing the sequence classification/regression loss.
+                Indices should be in :obj:`[0, ..., config.num_labels - 1]`.
+                If :obj:`config.num_labels == 1` a regression loss is computed (Mean-Square loss),
+                If :obj:`config.num_labels > 1` a classification loss is computed (Cross-Entropy).
+            output_attentions (`bool`, *optional*): Whether or not to return the attentions tensors of all attention layers.
+            output_hidden_states: not used.
+            event_tokens: a mask defining which tokens in the input are to be averaged for input to classifier head; only used when self.tokens==True.
+
+        Returns: (`transformers.SequenceClassifierOutput`) the output of the model
         """
 
-        outputs = self.encoder(
-            input_ids,
+        kwargs = self.generalize_encoder_forward_kwargs(
             attention_mask=attention_mask,
             token_type_ids=token_type_ids,
             position_ids=position_ids,
@@ -364,6 +451,11 @@ class CnlpModelForClassification(PreTrainedModel):
             output_hidden_states=True,
             return_dict=True
         )
+
+        outputs = self.encoder(
+            input_ids,
+            **kwargs
+        )
         
         batch_size,seq_len = input_ids.shape
 
@@ -402,4 +494,4 @@ class CnlpModelForClassification(PreTrainedModel):
                 loss=state['loss'], logits=logits, hidden_states=outputs.hidden_states, attentions=outputs.attentions,
             )
         else:
-            return SequenceClassifierOutput(loss=state['loss'], logits=logits)
+            return SequenceClassifierOutput(loss=state['loss'], logits=logits, attentions=outputs.attentions)
